#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "gkrellmbgchg.h"

static GtkWidget *gkrellm_vbox;

static Krell *krell_time;
static Panel *panel;
static Monitor *monitor;
static Decal *decal_wu;
static gint style_id;

static gint seconds_left;
static GList *idb;
static GtkTooltips *tip;
static gint locked;

/* config widgets */
static GtkWidget *entry_format_str;
static GtkWidget *entry_command, *entry_idb;
static GtkWidget *wait_seconds_spin_button;
static GtkWidget *randomise_entry;
static GtkWidget *reset_entry;

struct bg_monitor {
	gint wait_seconds;			/* sec. between updates */
	gint randomise;				/* randomise images? */
	gint reset;				/* reset the counter if lock is released */
	gchar format_string[64];	/* output format string */
	gchar command[256];			/* command to change bg */
	gchar idb[256];				/* full path to image database */
};

static struct bg_monitor bgmon = { 
	600, 1, 0,
	"$t",
	"Esetroot -f", 
	"/home/user/images.idb"
};

struct idb_entry {
	gchar *filename;
};

static void update_image_list(void) 
{
	gchar tmp[512];
	FILE *idb_file;
	struct idb_entry *idb_e;

	if(idb != NULL) { g_list_free(idb); idb = NULL; }
	
	if( (idb_file = fopen( bgmon.idb, "r" )) == NULL) {
		/* fwiw, print an error message to stderr */
		fprintf( stderr, "could not open image database.\n" );
		return;
	}

	while(!feof(idb_file))
	/* abort on error and eof when no bytes were read */
	if( fgets( tmp, 512, idb_file ) ) {
		if( tmp[strlen(tmp)-1]=='\n' )
			tmp[strlen(tmp)-1] = 0x00; /* strip trailing newline */
		if( *tmp && *tmp != '#') {
			/* only append, if it is really supposed to be a filename */
			idb_e = (struct idb_entry *)calloc( 1, sizeof( struct idb_entry ));
			idb_e->filename = g_strdup( tmp );
			idb = g_list_append( idb, idb_e );
		}
	}

	fclose(idb_file);
}

static void update_image(void)
{
	gchar *tmp = NULL;
	gchar *fname = NULL;
	gchar *tiptext = NULL;
	gint randval;

	if( g_list_length( idb )) {
		if( bgmon.randomise ) {
			srandom(time(NULL));
			randval = (double)random() / (double)RAND_MAX * (double)(g_list_length( idb ));
			fname = g_strdup( ((struct idb_entry *)g_list_nth( idb, randval )->data)->filename );
		} else {
			fname = g_strdup( ((struct idb_entry *)idb->data)->filename );
			if( idb->next ) idb = idb->next;
			else idb = g_list_first( idb );
		}
		tmp = g_strdup_printf( "%s '%s'", bgmon.command, fname );
		system( tmp );
		g_free( tmp );
	}

	if( tiptext == NULL && fname ) tiptext = g_strdup( fname );
	gtk_tooltips_set_tip( tip, panel->drawing_area, tiptext, NULL );
	gtk_tooltips_enable( tip );
	g_free( tiptext );
	g_free( fname );
}

static void update_decals_text( gchar *text )
{
	gchar *s, buf[48];

  if( !locked ) {
	text[0] = '\0';
	for( s=bgmon.format_string; *s!='\0'; s++ ) {
		buf[0] = *s; buf[1]='\0';
		if( *s == '$' && *(s+1) !='\0' ) {
			switch( *(s+1) ) {
				case 's':
					snprintf( buf, 12, "%d", seconds_left );
					s++;
					break;
				case 'S':
					snprintf( buf, 12, "%d", bgmon.wait_seconds - seconds_left );
					s++;
					break;
				case 'm':
					snprintf( buf, 12, "%d", seconds_left / 60 );
					s++;
					break;
				case 'M':
					snprintf( buf, 12, "%d", (bgmon.wait_seconds-seconds_left) / 60 );
					s++;
					break;
				case 't':
					snprintf( buf, 12, "-%.2d:%.2d",
							seconds_left / 60, seconds_left % 60 );
					s++;
					break;
				case 'T':
					snprintf( buf, 12, "%.2d:%.2d",
							(bgmon.wait_seconds-seconds_left) / 60,
							(bgmon.wait_seconds-seconds_left) % 60 );
					s++;
					break;
			}
		}
		strcat( text, buf );
	}
  }
}

static void update_krell(void)
{
	gkrellm_update_krell( panel, krell_time, (gulong) bgmon.wait_seconds - seconds_left );
}

static void update_plugin(void)
{
	gchar text[128] = "bgchg";

	if(GK.second_tick) {
		if( !locked )
			if( !(seconds_left--)) {
				update_image();
				seconds_left = bgmon.wait_seconds;
			}
	}

	if( locked ) {
		gkrellm_draw_decal_text( panel, decal_wu, "locked", -1 );
	} else {
		update_decals_text( text );
		gkrellm_draw_decal_text( panel, decal_wu, text, -1 );
	}
	update_krell();
	gkrellm_draw_layers( panel );
}

static gint panel_expose_event( GtkWidget *widget, GdkEventExpose *ev)
{
	if( widget == panel->drawing_area ) {
		gdk_draw_pixmap( widget->window,
						widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
						panel->pixmap,
						ev->area.x, ev->area.y,
						ev->area.x, ev->area.y,
						ev->area.width, ev->area.height );
	}
	return FALSE;
}

static gint cb_button_press( GtkWidget *widget, GdkEventButton *ev )
{
	gint shift_modifier = (ev->state & GDK_SHIFT_MASK ) ? 1 : 0;

	switch( ev->button ) {
		/* left */
		case 1:
			if( shift_modifier ) /* shift+left -> lock/unlock the bg */
				if( locked ) {
					locked = 0;
					if( bgmon.reset ) seconds_left = bgmon.wait_seconds;
				} else locked = 1;
			else {
				update_image();
				seconds_left = bgmon.wait_seconds;
			}
			break;
		/* middle */
		case 2:
			update_image_list();
			break;
		/* right */
		case 3:
			if( shift_modifier ) /* shift+right -> lock/unlock the bg */
				if( locked ) {
					locked = 0;
					if( bgmon.reset ) seconds_left = bgmon.wait_seconds;
				} else locked = 1;
			else
				gkrellm_open_config_window( monitor );
			break;
		/* wheel up */
		case 4:
			locked = 1;
			break;
		/* wheel down */
		case 5:
			if( locked ) {
				locked = 0;
				if( bgmon.reset ) seconds_left = bgmon.wait_seconds;
			}
	}

	return FALSE;
}

static void create_plugin( GtkWidget *vbox, gint first_create )
{
	GdkImlibImage *krell_img;
	Style *style;
	TextStyle *ts;
	gchar text[32] = "bgchg" ;

	gkrellm_vbox = vbox;

	if (first_create) panel = gkrellm_panel_new0 ();

	style = gkrellm_meter_style( style_id );
	krell_img = gkrellm_krell_meter_image( style_id );
	ts = gkrellm_panel_textstyle( style_id );
	panel->textstyle = ts;

	krell_time = gkrellm_create_krell( panel, krell_img, style );
	gkrellm_monotonic_krell_values(krell_time, FALSE);
	gkrellm_set_krell_full_scale( krell_time, bgmon.wait_seconds, 1 );

	decal_wu = gkrellm_create_decal_text( panel, "Apif0", ts, style, -1, -1, -1 );

	gkrellm_panel_configure( panel, NULL, style );
	gkrellm_panel_create( vbox, monitor, panel );

	gkrellm_draw_decal_text( panel, decal_wu, text, -1 );

	if(first_create) {
		gtk_signal_connect( GTK_OBJECT (panel->drawing_area), "expose_event",
				(GtkSignalFunc) panel_expose_event, NULL );
		gtk_signal_connect( GTK_OBJECT (panel->drawing_area), "button_press_event",
				(GtkSignalFunc) cb_button_press, NULL );
	}
	tip = gtk_tooltips_new();
	gtk_tooltips_enable( tip );

	locked = 0;
	srandom(time(NULL));
	update_image_list();
	seconds_left = bgmon.wait_seconds;
	update_image();
	update_krell();

	gkrellm_draw_layers( panel );
}

static gchar *plugin_info_text[] = {
	"<b>GKrellMBgChg ",
	"is a GKrellM plugin which periodically changes\n",
	"your desktop background. It also allows you to monitor the time\n",
	"between the updates in various ways. It is possible, to force a\n",
	"change by clicking on the panel.\n\n",
	"<b>Mouse Actions:\n\n",
	"<b>- Left ",
	"changes the background image and resets the timer,\n",
	"\t+<Shift> toggles the background lock.\n",
	"<b>- Middle ",
	"reloads the images from the image database (see Image Database).\n",
	"<b>- Right ",
	"opens the GKrellM Background Changer config window,\n",
	"\t+<Shift> toggles the background lock.\n\n",
	"<b>Mouse Wheel Actions:\n\n",
	"<b>- Up ",
	"\"locks\" the current background image (if you like it very much).\n",
	"<b>- Down ",
	"\"unlocks\" the image and the counter continues.\n\n",
	"<b>Configuration:\n\n",
	"<b>- Format String\n",
	"\tThe text output format is controlled by this string (default: $s).\n",
	"<b>\t$s ", "are the seconds that are remaining to the next update\n",
	"<b>\t$S ", "are the seconds that passed since the last change\n",
	"<b>\t$m ", "are the minutes that are remaining to the next update\n",
	"<b>\t$M ", "are the minutes that passed since the last change\n",
	"<b>\t$t ", "is the time remaining to the next update, displayed as '-mm:ss'\n",
	"<b>\t$T ", "is the time that passed since the last change, displayed as 'mm:ss'.\n\n",
	"<b>- Command\n",
	"\tProgram to use to set the desktop background image (including args),\n",
	"\tdefault: ",
	"<i>Esetroot -f\n\n",
	"<b>- Image Database\n",
	"\tfull path to a file containing all the images (full path/line) to be\n",
	"\tused by the plugin. (e.g. the output from ",
	"<i>find / -name *.jpg | sort",
	")\n\tEntries starting with a '#' will be ",
	"<b>ignored.\n",
	"\tdefault: ",
	"<i>/home/user/images.idb\n\n",
	"<b>- Randomise images\n",
	"\tSelect whether the image list should be randomised or not.\n",
	"<i>\tNote: ",
	"If it is not set, it will ",
	"<b>always ",
	"start at the first image in the list.\n\n",
	"<b>- Reset counter\n",
	"\tReset the counter to the initial value, if the \"image lock\" is released.\n"
};

static void create_bgchg_tab( GtkWidget *tab )
{
	GtkWidget *tabs, *vbox, *hbox;
	GtkWidget *label, *frame;
	GtkWidget *scrolled, *text;
	GtkAdjustment *adj;
	gchar *about_text = NULL;

	tabs = gtk_notebook_new();
	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(tabs),GTK_POS_TOP);
	gtk_box_pack_start(GTK_BOX(tab),tabs,TRUE,TRUE,0);  

	/* options */
	frame = gtk_frame_new(NULL);
	gtk_container_border_width(GTK_CONTAINER(frame),3);
	label = gtk_label_new("Options");
	gtk_notebook_append_page(GTK_NOTEBOOK(tabs),frame,label);  
 
	vbox = gtk_vbox_new(FALSE,0);
	gtk_container_add(GTK_CONTAINER(frame),vbox);

	hbox = gtk_hbox_new(FALSE, 0);
	label = gtk_label_new("Format String");
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 4);
	entry_format_str = gtk_entry_new_with_max_length(127);
	gtk_entry_set_text(GTK_ENTRY(entry_format_str),bgmon.format_string);
	gtk_box_pack_start(GTK_BOX(hbox), entry_format_str, FALSE, FALSE, 4);
	gtk_container_add(GTK_CONTAINER(vbox),hbox);

	hbox = gtk_hbox_new(FALSE, 0);
	label = gtk_label_new("Background Change Command");
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 4);
	entry_command = gtk_entry_new_with_max_length(127);
	gtk_entry_set_text(GTK_ENTRY(entry_command),bgmon.command);
	gtk_box_pack_start(GTK_BOX(hbox), entry_command, TRUE, TRUE, 4);
	gtk_container_add(GTK_CONTAINER(vbox),hbox);

	hbox = gtk_hbox_new(FALSE, 0);
	label = gtk_label_new("Image Database");
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 4);
	entry_idb = gtk_entry_new_with_max_length(127);
	gtk_entry_set_text(GTK_ENTRY(entry_idb),bgmon.idb);
	gtk_box_pack_start(GTK_BOX(hbox), entry_idb, TRUE, TRUE, 4);
	gtk_container_add(GTK_CONTAINER(vbox),hbox);

	hbox = gtk_hbox_new(FALSE, 0);
	adj = (GtkAdjustment *) gtk_adjustment_new(bgmon.wait_seconds,
												1.0, 3000.0, 1.0, 5.0, 0.0);
	wait_seconds_spin_button = gtk_spin_button_new(adj, 0.5, 0);
	gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(wait_seconds_spin_button), TRUE);
	gtk_box_pack_start(GTK_BOX(hbox), wait_seconds_spin_button, FALSE, FALSE, 4);
	label = gtk_label_new("Seconds between image changes");
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 4);
	gtk_container_add(GTK_CONTAINER(vbox),hbox);

	randomise_entry = gtk_check_button_new_with_label( "Randomise images");
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( randomise_entry ), bgmon.randomise );
	gtk_container_add( GTK_CONTAINER(vbox), randomise_entry );

	reset_entry = gtk_check_button_new_with_label( "Reset counter on \"lock\" release");
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( reset_entry ), bgmon.reset );
	gtk_container_add( GTK_CONTAINER(vbox), reset_entry );

	/* info */
	vbox = gkrellm_create_tab(tabs,"Info");
	scrolled = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled),
			GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_box_pack_start(GTK_BOX(vbox), scrolled, TRUE, TRUE, 0);
	text = gtk_text_new(NULL, NULL);
	gkrellm_add_info_text(text, plugin_info_text,
			sizeof(plugin_info_text) / sizeof(gchar *));
	gtk_text_set_editable(GTK_TEXT(text), FALSE);
	gtk_container_add(GTK_CONTAINER(scrolled), text);
	
	/* about */
	about_text = g_strdup_printf(
		"GKrellMBgChg %s\n" \
		"GKrellM Background Changer Plugin\n\n" \
		"Copyright (C) 2002 Stefan Bender\n" \
		"oraqhf@gmx.net\n" \
		"http://www.uni-jena.de/~p6best/english/comp/gkrellmbgchg.html\n\n" \
		"Released under the GNU Public Licence",
		GKRELLMBGCHG_VERSION);
	
	text = gtk_label_new(about_text); 
	label = gtk_label_new("About");
	gtk_notebook_append_page(GTK_NOTEBOOK(tabs),text,label);
	g_free(about_text);
}

static void apply_config(void)
{
	gchar *s;

	/* update config vars */
	bgmon.wait_seconds = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(wait_seconds_spin_button));
	s = gtk_entry_get_text(GTK_ENTRY(entry_format_str));
	strcpy(bgmon.format_string,s);
	s = gtk_entry_get_text(GTK_ENTRY(entry_command));
	strcpy(bgmon.command,s);
	s = gtk_entry_get_text(GTK_ENTRY(entry_idb));
	strcpy(bgmon.idb,s);
	bgmon.randomise = GTK_TOGGLE_BUTTON( randomise_entry )->active;
	bgmon.reset = GTK_TOGGLE_BUTTON( reset_entry )->active;

	/* delete old panel */
	if(panel != NULL)
	{
		gkrellm_panel_destroy(panel);
		panel = NULL;
	}
	/* create new panel */
	create_plugin(gkrellm_vbox,1);
}

static void save_config( FILE *f)
{
	fprintf( f, "%s wait_seconds %d\n", CONFIG_KEYWORD, bgmon.wait_seconds );
	fprintf( f, "%s randomise %d\n", CONFIG_KEYWORD, bgmon.randomise );
	fprintf( f, "%s reset %d\n", CONFIG_KEYWORD, bgmon.reset );
	fprintf( f, "%s format_string %s\n", CONFIG_KEYWORD, bgmon.format_string );
	fprintf( f, "%s command %s\n", CONFIG_KEYWORD, bgmon.command );
	fprintf( f, "%s idb %s\n", CONFIG_KEYWORD, bgmon.idb );
}

static void load_config( gchar *arg )
{
	gchar *command, *p;

	for( p = arg; *p && isspace(*p); p++ );
	for( ; *p && !isspace(*p); p++ );

	command = malloc( (p-arg+1)*sizeof(char) );
	memset( command, '\0', p-arg+1 );
	memcpy( command, arg, p-arg );

	for( ; *p && isspace(*p); p++ );

	if( !strcmp(command, "wait_seconds") ) bgmon.wait_seconds = atoi( p );
	else if( !strcmp(command, "randomise") ) bgmon.randomise = atoi( p );
	else if( !strcmp(command, "reset") ) bgmon.reset = atoi( p );
	else if( !strcmp(command, "format_string") ) strcpy( bgmon.format_string, p );
	else if( !strcmp(command, "command") ) strcpy( bgmon.command, p );
	else if( !strcmp(command, "idb") ) strcpy( bgmon.idb, p );

	free( command );
}

static Monitor plugin_mon = {
	CONFIG_NAME,
	0,
	create_plugin,
	update_plugin,
	create_bgchg_tab,
	apply_config,
	save_config,
	load_config,
	CONFIG_KEYWORD,
	NULL,
	NULL,
	NULL,
	PLACEMENT,
	NULL,
	NULL
};

Monitor *init_plugin()
{
	style_id = gkrellm_add_meter_style( &plugin_mon, STYLE_NAME );
	monitor = &plugin_mon;
	return &plugin_mon;
}
